from random import sample
from collections import defaultdict
from functools import partial
from multiprocessing import Pool, cpu_count
from dataclasses import dataclass
from typing import Union, List, Tuple, Dict

from krippendorff import alpha

from cml.domain.data_source import DataSource
from cml.shared.settings import ReconstructionSettings
from cml.shared.errors import (
    NoModelReconstructedError,
    NotEnoughFeaturesWarning)

__all__ = (
    "Reconstructor",
)


@dataclass
class Metadata:
    knowledge_domain: str
    knowledge_tier: int
    identifier: int
    pre_image: list
    pre_image_features: list
    pre_image_labels: list
    t_min: int
    t_max: int
    sigma: tuple
    zeta: tuple

    def __str__(self):
        return f"Knowledge domain: <{self.knowledge_domain}> \n" \
               f"Knowledge tier: <{self.knowledge_tier}> \n" \
               f"Identifier: <{self.identifier}> \n" \
               f"Pre image: <{self.pre_image}> \n" \
               f"Pre image labels: <{self.pre_image_labels}> \n" \
               f"Pre image features: <{self.pre_image_features}> \n" \
               f"T min: <{self.t_min}> \n" \
               f"T max: <{self.t_max}> \n" \
               f"Subjects: <{self.sigma}> \n" \
               f"Purposes: <{self.zeta}> \n"

    def __hash__(self):
        return hash(".".join([self.knowledge_domain,
                              str(self.knowledge_tier),
                              str(self.identifier)]))

    def __eq__(self, other):
        if not isinstance(other, Metadata):
            raise TypeError()

        return hash(self) == hash(other)


class PragmaticMachineLearningModel:
    def __init__(self,
                 meta: Metadata,
                 model,
                 learnblock):
        self.meta = meta
        self.model = model
        self.origin = learnblock.origin
        self.learnblock = learnblock if self.tier > 1 else None

    def __str__(self) -> str:
        return self.uid

    def __repr__(self) -> str:
        return self.uid

    def __hash__(self) -> int:
        return hash(self.meta)

    def __eq__(self,
               other: Union['PragmaticMachineLearningModel', str]) -> bool:
        if isinstance(other, PragmaticMachineLearningModel):
            return hash(self) == hash(other)

        if isinstance(other, str):
            return hash(self) == hash(other)

        raise NotImplementedError()

    @property
    def pre_image_features(self) -> List[str]:
        return self.meta.pre_image_features

    @property
    def tier(self) -> int:
        return self.meta.knowledge_tier

    @property
    def min_timestamp(self) -> int:
        return self.meta.t_min

    @property
    def max_timestamp(self) -> int:
        return self.meta.t_max

    @property
    def pre_image(self) -> List[int]:
        return self.meta.pre_image

    @property
    def pre_image_labels(self) -> List[Union[int, str]]:
        return self.meta.pre_image_labels

    @property
    def subject(self) -> Tuple[str]:
        return self.meta.sigma

    @property
    def purpose(self) -> Tuple[str]:
        return self.meta.zeta

    @property
    def uid(self) -> str:
        return ".".join([self.meta.knowledge_domain,
                         str(self.meta.knowledge_tier),
                         str(self.meta.identifier)])

    @property
    def identifier(self):
        return self.meta.identifier

    @identifier.setter
    def identifier(self, value):
        self.meta.identifier = value

    @property
    def counter(self):
        return self.meta.identifier

    def fusion(self,
               model: 'PragmaticMachineLearningModel',
               knowledge_tier: int,
               new_identifier: int) -> Metadata:

        return Metadata(self.meta.knowledge_domain,
                        knowledge_tier,
                        new_identifier,
                        None,
                        None,
                        None,
                        min(self.meta.t_min, model.min_timestamp),
                        max(self.meta.t_max, model.max_timestamp),
                        tuple(set(self.subject + model.subject)),
                        tuple(set(self.meta.zeta + model.purpose)))

        # return Metadata(self.meta.knowledge_domain,
        #                 self.meta.knowledge_tier,
        #                 new_identifier,
        #                 self.pre_image + model.pre_image,
        #                 list(set(self.pre_image_features).intersection(
        #                      set(model.pre_image_features))),
        #                 self.pre_image_labels + model.pre_image_labels,
        #                 min(self.meta.t_min, model.min_timestamp),
        #                 max(self.meta.t_max, model.max_timestamp),
        #                 tuple(set(self.subject + model.subject)),
        #                 tuple(set(self.meta.zeta + model.purpose)))
        #

    def trained_with(self, source: DataSource):
        if self.origin == "source":
            block = source.get_block(self.pre_image,
                                     columns=self.pre_image_features)
            block = block.set_labels(self.pre_image_labels)
            return block

        else:
            return self.learnblock


def log(func):
    def wrapper(self, tier, learnblock, which_models=None, meta=None):
        self.logger.protocol("{:*^100}".format("Reconstruction"))
        self.logger.protocol(self.settings)
        return func(self, tier, learnblock, which_models, meta)
    return wrapper


class SubprocessDTO:
    def __init__(self, model, learnblock, tier, abv, settings):
        self.model = model
        self.learnblock = learnblock
        self.tier = tier
        self.abv = abv
        self.settings = settings


class Reconstructor:
    CONCEPTUAL_KNOWLEDGE_ABBREVIATION = "C"
    PROCEDURAL_KNOWLEDGE_ABBREVIATION = "P"

    def __init__(self,
                 settings: ReconstructionSettings,
                 ml_models: List,
                 mode: str):
        self.mode = mode
        self.settings = settings
        self.ml_models = ml_models
        self.__reconstruction = self._init_reconstruction(mode)
        self.next_model_counter = None
        self.logger = None

    def _init_reconstruction(self, mode: str):
        return partial(self._reconstruct_conceptual, krippen="nominal") \
               if mode == "conceptual" else \
               partial(self._reconstruct_procedural, krippen="ratio")

    def reconstruct_parallel(self, tier, learnblock, models=None, meta=None):
        self.logger.protocol("{:^100}".format("Starting Reconstruction"))
        if not models:
            args = []
            for model in self.ml_models:
                args.append(
                    SubprocessDTO(model, learnblock, tier,
                                  self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                                  self.settings)
                )
            pool = Pool(cpu_count()-2)
            self.logger.protocol("{:^100}".format("Mapped"))
            reliability_model = pool.map(reconstruct_in_parallel, args)
            pool.close()
            pool.join()
            self.logger.protocol("{:^100}".format("Result erhalten"))
            reliabilities_to_model = defaultdict(list)
            for r, m in reliability_model:
                if r and m:
                    reliabilities_to_model[r].append(m)
            if reliabilities_to_model.keys():
                winner = determine_winner(reliabilities_to_model)
                winner.identifier = self.next_model_counter(tier)
                return winner
            raise NoModelReconstructedError()

    @log
    def reconstruct(self,
                    tier: int,
                    learnblock,
                    which_models: List = None,
                    meta: Metadata = None) -> PragmaticMachineLearningModel:

        # Check if learnblock has enough features
        if not (learnblock.learn_cols >= 2):
            raise NotEnoughFeaturesWarning()

        # Specify the models which should be trained
        if not which_models:
            which_models = [m.subject for m in self.ml_models]

        # Start the reconstruction
        reliabilities_to_model = self.__reconstruction(tier,
                                                       learnblock,
                                                       which_models,
                                                       meta=meta)

        # Determine the best pragmatic machine learning model
        if reliabilities_to_model.keys():
            return determine_winner(reliabilities_to_model)
        raise NoModelReconstructedError()

    def _reconstruct_conceptual(self,
                                tier: int,
                                learnblock,
                                which_models: List,
                                krippen: str = None,
                                meta: Metadata = None):

        reliability_to_model = defaultdict(list)
        for model in self.ml_models:
            if model.subject not in which_models: continue

            # train model
            train_block, eval_block = split(learnblock, self.settings.reliability_sample)
            try:
                trained_model = model.train(
                    train_block.as_numpy_array(),
                    [i for i in train_block.get_column_values("Z")])
            except TypeError as error:
                print(train_block.purpose)
                print(error.with_traceback())
                input()
            ###################################################################
            self.logger.protocol("\n{:<20}".format(model.subject))
            self.logger.protocol("{:<20}: {}".format("Accuracy", model.accuracy))
            ###################################################################

            # check constraints
            if valid_reconstructed(trained_model,
                                   "conceptual",
                                   self.settings.min_test_accuracy,
                                   self.settings.max_test_error_avg,
                                   self.settings.max_test_error_max):
                reliability = calc_reliability(trained_model, learnblock, krippen)
                ###################################################################
                self.logger.protocol("{:<20}: {}".format("Reliability", str(reliability)))
                ###################################################################
                if reliability >= self.settings.min_reliability:
                    if not meta:
                        uid = ".".join([self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                                        str(tier),
                                        learnblock.purpose])
                        meta = Metadata(self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                                        tier,
                                        self.next_model_counter(tier),
                                        learnblock.indexes,
                                        learnblock.columns(),
                                        learnblock.get_column_values("Z"),
                                        learnblock.min_timestamp,
                                        learnblock.max_timestamp,
                                        (model.subject, ),
                                        (uid, ))
                    else:
                        meta.pre_image_features = learnblock.columns()
                        meta.pre_image_labels = learnblock.get_column_values("Z")
                        meta.pre_image = learnblock.indexes

                    reliability_to_model[reliability].append(
                        PragmaticMachineLearningModel(
                            meta, trained_model, learnblock
                        )
                    )
        return reliability_to_model

    def _reconstruct_procedural(self,
                                tier: int,
                                learnblock,
                                which_models: List,
                                krippen: str = None,
                                meta: Metadata = None) -> Dict[float, List]:
        reliability_to_model = defaultdict(list)
        for model in self.ml_models:
            if model.subject not in which_models: continue

            # train model
            train_block, eval_block = self.split(learnblock)
            trained_model = model.train(
                train_block.as_numpy_array(),
                [i for i in train_block.get_column_values("Z")])

            # check contraints
            if valid_reconstructed(trained_model, "procedural"):
                reliability = self.calc_reliability(
                    trained_model, learnblock, krippen)
                if reliability >= self.settings.min_reliability:
                    if not meta:
                        prag_meta_data = Metadata(
                            "P",
                            tier,
                            next(self.next_model_counter),
                            learnblock.indexes,
                            learnblock.columns(),
                            learnblock.get_column_values("Z"),
                            learnblock.min_timestamp,
                            learnblock.max_timestamp,
                            (model.subject, ),
                            (".".join(["C", '1', learnblock.purpose]), )
                        )
                    else:
                        meta.pre_image_features = learnblock.columns()
                        meta.pre_image_labels = learnblock.get_column_values("Z")
                        meta.pre_image = learnblock.indexes

                    reliability_to_model[reliability].append(
                        PragmaticMachineLearningModel(prag_meta_data,
                                                      trained_model,
                                                      learnblock))
        return reliability_to_model


def split(learnblock, reliability_sample) -> Tuple:
    indices = learnblock.indexes
    eval_size = int(learnblock.length * reliability_sample)
    eval_idx = sample(indices, eval_size)
    train_idx = list(set(indices).difference(set(eval_idx)))
    return learnblock.new_block_from_rows_index(train_idx), \
        learnblock.new_block_from_rows_index(eval_idx)


def calc_reliability(trained_model, eval_block, metric: str) -> float:
    y_pre = trained_model.predict(eval_block.as_numpy_array())
    y_true = [i for i in eval_block.get_column_values("Z")]
    reliability_data = [y_pre, y_true]
    return alpha(reliability_data, level_of_measurement=metric)


def valid_reconstructed(model,
                        knowledge_domain: str,
                        min_test_accuracy,
                        max_test_error_avg,
                        max_test_error_max) -> bool:
    if knowledge_domain == "conceptual":
        return model.accuracy >= min_test_accuracy
    else:
        return model.mean_error <= max_test_error_avg and \
            model.max_error <= max_test_error_max


def reconstruct_in_parallel(*args):
    dto = args[0]
    train_block, eval_block = split(dto.learnblock, dto.settings.reliability_sample)
    trained_model = dto.model.train(train_block.as_numpy_array(),
                                [i for i in train_block.get_column_values("Z")])
    if valid_reconstructed(trained_model,
                           "conceptual",
                           dto.settings.min_test_accuracy,
                           dto.settings.max_test_error_avg,
                           dto.settings.max_test_error_max):
        reliability = calc_reliability(trained_model, dto.learnblock, "nominal")
        if reliability >= dto.settings.min_reliability:
                uid = ".".join([dto.abv,
                                str(dto.tier),
                                dto.learnblock.purpose])
                meta = Metadata(dto.abv,
                                dto.tier,
                                0,
                                dto.learnblock.indexes,
                                dto.learnblock.columns(),
                                dto.learnblock.get_column_values("Z"),
                                dto.learnblock.min_timestamp,
                                dto.learnblock.max_timestamp,
                                (dto.model.subject,),
                                (uid,))
                return reliability, PragmaticMachineLearningModel(
                    meta, trained_model, dto.learnblock)

    return None, None


def determine_winner(
        reliability_to_model: dict) -> PragmaticMachineLearningModel:
    sorted_reliabilities = sorted(reliability_to_model.keys(), reverse=True)
    biggest_reliabilities = reliability_to_model[sorted_reliabilities.pop()]
    winner = None
    min_domain = float("inf")
    for model in biggest_reliabilities:
        if len(model.pre_image_features) < min_domain:
            min_domain = len(model.pre_image_features)
            winner = model

    return winner
